@@ -69,7 +69,7 @@
 #define  STDERR  2
 
 #ifndef HAVE_SOCKLEN_T
-#define socklen_t int
+typedef int socklen_t;
 #endif
 
 #define GKD_COMP_KEYRING    "keyring"
@@ -775,7 +775,6 @@ on_login_timeout (gpointer data)
 int
 main (int argc, char *argv[])
 {
-	GMainContext *ctx;
 	GMainLoop *loop;
 
 	/*
@@ -804,7 +803,7 @@ main (int argc, char *argv[])
 	 */
 	gkd_capability_obtain_capability_and_drop_privileges ();
 
-#ifdef WITH_TESTS
+#ifdef WITH_TESTABLE
 	g_setenv ("DBUS_FATAL_WARNINGS", "1", FALSE);
 	if (!g_getenv ("G_DEBUG"))
 		g_log_set_always_fatal (G_LOG_LEVEL_CRITICAL | G_LOG_LEVEL_WARNING);
@@ -835,7 +834,7 @@ main (int argc, char *argv[])
 	if (run_version) {
 		g_print ("gnome-keyring-daemon: %s\n", VERSION);
 		g_print ("testing: %s\n",
-#ifdef WITH_TESTS
+#ifdef WITH_TESTABLE
 		         "enabled");
 #else
 		         "disabled");
@@ -868,7 +867,6 @@ main (int argc, char *argv[])
 
 	/* Initialize our daemon main loop and threading */
 	loop = g_main_loop_new (NULL, FALSE);
-	ctx = g_main_loop_get_context (loop);
 
 	/* Initialize our control socket */
 	if (!gkd_control_listen ())
