@@ -9,6 +9,7 @@ const ASN1_ARRAY_TYPE test_asn1_tab[] = {
   { NULL, 1073741836, NULL },
   { "TestInteger", 1073741827, NULL },
   { "TestBoolean", 1073741828, NULL },
+  { "TestNull", 1073741844, NULL },
   { "TestOctetString", 1073741831, NULL },
   { "TestGeneralized", 1082130449, NULL },
   { "TestImplicit", 1610620935, NULL },
@@ -27,7 +28,24 @@ const ASN1_ARRAY_TYPE test_asn1_tab[] = {
   { NULL, 131081, NULL },
   { "TestOid", 1610612741, NULL },
   { "oid", 12, NULL },
-  { "TestAnySeq", 536870917, NULL },
+  { "TestAnySeq", 1610612741, NULL },
   { "contents", 13, NULL },
+  { "TestAnyExp", 1610612741, NULL },
+  { "contents", 536879117, NULL },
+  { NULL, 2056, "89"},
+  { "TestAnyChoice", 1610612754, NULL },
+  { "choiceShortTag", 1610620941, NULL },
+  { NULL, 2056, "30"},
+  { "choiceLongTag", 536879117, NULL },
+  { NULL, 2056, "31"},
+  { "TestSeqOf", 1610612747, NULL },
+  { NULL, 3, NULL },
+  { "TestSetOf", 1610612751, NULL },
+  { NULL, 3, NULL },
+  { "TestEnumerated", 537133077, NULL },
+  { "valueZero", 1073741825, "0"},
+  { "valueOne", 1073741825, "1"},
+  { "valueTwo", 1073741825, "2"},
+  { "valueThree", 1, "3"},
   { NULL, 0, NULL }
 };
