@@ -20,8 +20,11 @@
 #include "config.h"
 
 #include "gcr-certificate.h"
+#include "gcr-certificate-exporter.h"
+#include "gcr-certificate-extensions.h"
 #include "gcr-certificate-renderer.h"
 #include "gcr-display-view.h"
+#include "gcr-fingerprint.h"
 #include "gcr-icons.h"
 #include "gcr-simple-certificate.h"
 #include "gcr-renderer.h"
@@ -39,7 +42,6 @@
 
 /**
  * GcrCertificateRenderer:
- * @parent: The parent object
  *
  * An implementation of #GcrRenderer which renders certificates.
  */
@@ -59,43 +61,218 @@ enum {
 };
 
 struct _GcrCertificateRendererPrivate {
-	GcrCertificate *certificate;
-	GckAttributes *attributes;
+	GcrCertificate *opt_cert;
+	GckAttributes *opt_attrs;
 	guint key_size;
 	gchar *label;
 };
 
 static void gcr_renderer_iface_init (GcrRendererIface *iface);
+static void gcr_renderer_certificate_iface_init (GcrCertificateIface *iface);
 
 G_DEFINE_TYPE_WITH_CODE (GcrCertificateRenderer, gcr_certificate_renderer, G_TYPE_OBJECT,
-                         G_IMPLEMENT_INTERFACE (GCR_TYPE_RENDERER, gcr_renderer_iface_init));
+	G_IMPLEMENT_INTERFACE (GCR_TYPE_RENDERER, gcr_renderer_iface_init);
+	GCR_CERTIFICATE_MIXIN_IMPLEMENT_COMPARABLE ();
+	G_IMPLEMENT_INTERFACE (GCR_TYPE_CERTIFICATE, gcr_renderer_certificate_iface_init);
+);
+
+static GQuark OID_BASIC_CONSTRAINTS = 0;
+static GQuark OID_EXTENDED_KEY_USAGE = 0;
+static GQuark OID_SUBJECT_KEY_IDENTIFIER = 0;
+static GQuark OID_KEY_USAGE = 0;
+static GQuark OID_SUBJECT_ALT_NAME = 0;
 
 /* -----------------------------------------------------------------------------
  * INTERNAL
  */
 
 static gchar*
-calculate_label (GcrCertificateRenderer *self, GNode *asn)
+calculate_label (GcrCertificateRenderer *self)
 {
 	gchar *label;
 
 	if (self->pv->label)
 		return g_strdup (self->pv->label);
 
-	if (self->pv->attributes) {
-		if (gck_attributes_find_string (self->pv->attributes, CKA_LABEL, &label))
+	if (self->pv->opt_attrs) {
+		if (gck_attributes_find_string (self->pv->opt_attrs, CKA_LABEL, &label))
 			return label;
 	}
 
-	if (asn != NULL) {
-		label = egg_dn_read_part (egg_asn1x_node (asn, "tbsCertificate", "subject", "rdnSequence", NULL), "CN");
-		if (label != NULL)
-			return label;
-	}
+	label = gcr_certificate_get_subject_cn (GCR_CERTIFICATE (self));
+	if (label != NULL)
+		return label;
 
 	return g_strdup (_("Certificate"));
 }
 
+static gboolean
+append_extension_basic_constraints (GcrCertificateRenderer *self, GcrDisplayView *view,
+                                    gconstpointer data, gsize n_data)
+{
+	GcrRenderer *renderer = GCR_RENDERER (self);
+	gboolean is_ca = FALSE;
+	gint path_len = -1;
+	gchar *number;
+
+	if (!_gcr_certificate_extension_basic_constraints (data, n_data, &is_ca, &path_len))
+		return FALSE;
+
+	_gcr_display_view_append_heading (view, renderer, _("Basic Constraints"));
+
+	_gcr_display_view_append_value (view, renderer, _("Certificate Authority"),
+	                                is_ca ? _("Yes") : _("No"), FALSE);
+
+	number = g_strdup_printf ("%d", path_len);
+	_gcr_display_view_append_value (view, renderer, _("Max Path Length"),
+	                                path_len < 0 ? _("Unlimited") : number, FALSE);
+	g_free (number);
+
+	return TRUE;
+}
+
+static gboolean
+append_extension_extended_key_usage (GcrCertificateRenderer *self, GcrDisplayView *view,
+                                     gconstpointer data, gsize n_data)
+{
+	GcrRenderer *renderer = GCR_RENDERER (self);
+	GQuark *oids;
+	GString *text;
+	guint i;
+
+	oids = _gcr_certificate_extension_extended_key_usage (data, n_data);
+	if (oids == NULL)
+		return FALSE;
+
+	_gcr_display_view_append_heading (view, renderer, _("Extended Key Usage"));
+
+	text = g_string_new ("");
+	for (i = 0; oids[i] != 0; i++) {
+		if (i > 0)
+			g_string_append_unichar (text, GCR_DISPLAY_VIEW_LINE_BREAK);
+		g_string_append (text, egg_oid_get_description (oids[i]));
+	}
+
+	g_free (oids);
+
+	_gcr_display_view_append_value (view, renderer, _("Allowed Purposes"),
+	                                text->str, FALSE);
+
+	g_string_free (text, TRUE);
+
+	return TRUE;
+}
+
+static gboolean
+append_extension_subject_key_identifier (GcrCertificateRenderer *self, GcrDisplayView *view,
+                                         gconstpointer data, gsize n_data)
+{
+	GcrRenderer *renderer = GCR_RENDERER (self);
+	gpointer keyid;
+	gsize n_keyid;
+
+	keyid = _gcr_certificate_extension_subject_key_identifier (data, n_data, &n_keyid);
+	if (keyid == NULL)
+		return FALSE;
+
+	_gcr_display_view_append_heading (view, renderer, _("Subject Key Identifier"));
+	_gcr_display_view_append_hex (view, renderer, _("Key Identifier"), keyid, n_keyid);
+
+	g_free (keyid);
+
+	return TRUE;
+}
+
+static const struct {
+	guint usage;
+	const gchar *description;
+} usage_descriptions[] = {
+	{ GCR_KEY_USAGE_DIGITAL_SIGNATURE, N_("Digital signature") },
+	{ GCR_KEY_USAGE_KEY_ENCIPHERMENT, N_("Key encipherment") },
+	{ GCR_KEY_USAGE_DATA_ENCIPHERMENT, N_("Data encipherment") },
+	{ GCR_KEY_USAGE_KEY_AGREEMENT, N_("Key agreement") },
+	{ GCR_KEY_USAGE_KEY_CERT_SIGN, N_("Certificate signature") },
+	{ GCR_KEY_USAGE_CRL_SIGN, N_("Revocation list signature") }
+};
+
+static gboolean
+append_extension_key_usage (GcrCertificateRenderer *self, GcrDisplayView *view,
+                            gconstpointer data, gsize n_data)
+{
+	GcrRenderer *renderer = GCR_RENDERER (self);
+	gulong key_usage;
+	GString *text;
+	guint i;
+
+	if (!_gcr_certificate_extension_key_usage (data, n_data, &key_usage))
+		return FALSE;
+
+	text = g_string_new ("");
+
+	for (i = 0; i < G_N_ELEMENTS (usage_descriptions); i++) {
+		if (key_usage & usage_descriptions[i].usage) {
+			if (text->len > 0)
+				g_string_append_unichar (text, GCR_DISPLAY_VIEW_LINE_BREAK);
+			g_string_append (text, gettext (usage_descriptions[i].description));
+		}
+	}
+
+	_gcr_display_view_append_heading (view, renderer, _("Key Usage"));
+	_gcr_display_view_append_value (view, renderer, _("Usages"), text->str, FALSE);
+
+	g_string_free (text, TRUE);
+
+	return TRUE;
+}
+
+static gboolean
+append_extension_subject_alt_name (GcrCertificateRenderer *self, GcrDisplayView *view,
+                                   gconstpointer data, gsize n_data)
+{
+	GcrRenderer *renderer = GCR_RENDERER (self);
+	GArray *general_names;
+	GcrGeneralName *general;
+	guint i;
+
+	general_names = _gcr_certificate_extension_subject_alt_name (data, n_data);
+	if (general_names == NULL)
+		return FALSE;
+
+	_gcr_display_view_append_heading (view, renderer, _("Subject Alternative Names"));
+
+	for (i = 0; i < general_names->len; i++) {
+		general = &g_array_index (general_names, GcrGeneralName, i);
+		if (general->display == NULL)
+			_gcr_display_view_append_hex (view, renderer, general->description,
+			                              general->raw, general->n_raw);
+		else
+			_gcr_display_view_append_value (view, renderer, general->description,
+			                                general->display, FALSE);
+	}
+
+	_gcr_general_names_free (general_names);
+
+	return TRUE;
+}
+
+
+static gboolean
+append_extension_hex (GcrCertificateRenderer *self, GcrDisplayView *view,
+                      GQuark oid, gconstpointer data, gsize n_data)
+{
+	GcrRenderer *renderer = GCR_RENDERER (self);
+	const gchar *text;
+
+	_gcr_display_view_append_heading (view, renderer, _("Extension"));
+
+	/* Extension type */
+	text = egg_oid_get_description (oid);
+	_gcr_display_view_append_value (view, renderer, _("Identifier"), text, FALSE);
+	_gcr_display_view_append_hex (view, renderer, _("Value"), data, n_data);
+
+	return TRUE;
+}
+
 static gboolean
 append_extension (GcrCertificateRenderer *self, GcrDisplayView *view,
                   GNode *asn, const guchar *data, gsize n_data, gint index)
@@ -103,11 +280,10 @@ append_extension (GcrCertificateRenderer *self, GcrDisplayView *view,
 	GcrRenderer *renderer = GCR_RENDERER (self);
 	GNode *node;
 	GQuark oid;
-	gchar *display;
 	gsize n_value;
 	const guchar *value;
-	const gchar *text;
 	gboolean critical;
+	gboolean ret = FALSE;
 
 	/* Make sure it is present */
 	node = egg_asn1x_node (asn, "tbsCertificate", "extensions", index, NULL);
@@ -118,28 +294,32 @@ append_extension (GcrCertificateRenderer *self, GcrDisplayView *view,
 	oid = egg_asn1x_get_oid_as_quark (egg_asn1x_node (node, "extnID", NULL));
 	g_return_val_if_fail (oid, FALSE);
 
-	_gcr_display_view_append_heading (view, renderer, _("Extension"));
-
-
-	/* Extension type */
-	text = egg_oid_get_description (oid);
-	_gcr_display_view_append_value (view, renderer, _("Identifier"), text, FALSE);
-
-
 	/* Extension value */
 	value = egg_asn1x_get_raw_value (egg_asn1x_node (node, "extnValue", NULL), &n_value);
 
-	/* TODO: Parsing of extensions that we understand */
-	display = egg_hex_encode_full (value, n_value, TRUE, ' ', 1);
-	_gcr_display_view_append_value (view, renderer, _("Value"), display, TRUE);
-	g_free (display);
+	/* The custom parsers */
+	if (oid == OID_BASIC_CONSTRAINTS)
+		ret = append_extension_basic_constraints (self, view, value, n_value);
+	else if (oid == OID_EXTENDED_KEY_USAGE)
+		ret = append_extension_extended_key_usage (self, view, value, n_value);
+	else if (oid == OID_SUBJECT_KEY_IDENTIFIER)
+		ret = append_extension_subject_key_identifier (self, view, value, n_value);
+	else if (oid == OID_KEY_USAGE)
+		ret = append_extension_key_usage (self, view, value, n_value);
+	else if (oid == OID_SUBJECT_ALT_NAME)
+		ret = append_extension_subject_alt_name (self, view, value, n_value);
 
+	/* Otherwise the default raw display */
+	if (ret == FALSE)
+		ret = append_extension_hex (self, view, oid, value, n_value);
 
 	/* Critical */
-	if (egg_asn1x_get_boolean (egg_asn1x_node (node, "critical", NULL), &critical))
-		_gcr_display_view_append_value (view, renderer, _("Critical"), critical ? _("Yes") : _("No"), FALSE);
+	if (ret == TRUE && egg_asn1x_get_boolean (egg_asn1x_node (node, "critical", NULL), &critical)) {
+		_gcr_display_view_append_value (view, renderer, _("Critical"),
+		                                critical ? _("Yes") : _("No"), FALSE);
+	}
 
-	return TRUE;
+	return ret;
 }
 
 typedef struct _on_parsed_dn_args {
@@ -188,6 +368,62 @@ on_parsed_dn_part (guint index, GQuark oid, const guchar *value,
 	g_free (display);
 }
 
+static gboolean
+on_delete_unref_dialog (GtkWidget *widget, GdkEvent *event, gpointer data)
+{
+	g_object_unref (widget);
+	return FALSE;
+}
+
+static void
+on_export_completed (GObject *source, GAsyncResult *result, gpointer user_data)
+{
+	GtkWindow *parent = GTK_WINDOW (user_data);
+	GcrCertificateExporter *exporter = GCR_CERTIFICATE_EXPORTER (source);
+	GError *error = NULL;
+	GtkWidget *dialog;
+
+	if (!_gcr_certificate_exporter_export_finish (exporter, result, &error)) {
+		if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED)) {
+			dialog = gtk_message_dialog_new_with_markup (parent,
+				  GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_ERROR,
+				  GTK_BUTTONS_OK, "<big>%s</big>\n\n%s",
+				  _("Couldn't export the certificate."),
+				  error->message);
+			gtk_widget_show (dialog);
+			g_signal_connect (dialog, "delete-event",
+					  G_CALLBACK (on_delete_unref_dialog), NULL);
+		}
+	}
+
+	/* Matches ref in on_certificate_export */
+	if (parent)
+		g_object_unref (parent);
+}
+
+static void
+on_certificate_export (GtkMenuItem *menuitem, gpointer user_data)
+{
+	GcrCertificateRenderer *self = GCR_CERTIFICATE_RENDERER (user_data);
+	GcrCertificateExporter *exporter;
+	gchar *label;
+	GtkWidget *parent;
+
+	label = calculate_label (self);
+
+	parent = gtk_widget_get_toplevel (GTK_WIDGET (menuitem));
+	if (parent && !GTK_IS_WINDOW (parent))
+		parent = NULL;
+
+	exporter = _gcr_certificate_exporter_new (GCR_CERTIFICATE (self), label,
+	                                          GTK_WINDOW (parent));
+
+	g_free (label);
+
+	_gcr_certificate_exporter_export_async (exporter, NULL, on_export_completed,
+	                                        parent ? g_object_ref (parent) : NULL);
+}
+
 /* -----------------------------------------------------------------------------
  * OBJECT
  */
@@ -203,9 +439,9 @@ gcr_certificate_renderer_dispose (GObject *obj)
 {
 	GcrCertificateRenderer *self = GCR_CERTIFICATE_RENDERER (obj);
 
-	if (self->pv->certificate)
-		g_object_unref (self->pv->certificate);
-	self->pv->certificate = NULL;
+	if (self->pv->opt_cert)
+		g_object_unref (self->pv->opt_cert);
+	self->pv->opt_cert = NULL;
 
 	G_OBJECT_CLASS (gcr_certificate_renderer_parent_class)->dispose (obj);
 }
@@ -215,11 +451,11 @@ gcr_certificate_renderer_finalize (GObject *obj)
 {
 	GcrCertificateRenderer *self = GCR_CERTIFICATE_RENDERER (obj);
 
-	g_assert (!self->pv->certificate);
+	g_assert (!self->pv->opt_cert);
 
-	if (self->pv->attributes)
-		gck_attributes_unref (self->pv->attributes);
-	self->pv->attributes = NULL;
+	if (self->pv->opt_attrs)
+		gck_attributes_unref (self->pv->opt_attrs);
+	self->pv->opt_attrs = NULL;
 
 	g_free (self->pv->label);
 	self->pv->label = NULL;
@@ -260,16 +496,16 @@ gcr_certificate_renderer_get_property (GObject *obj, guint prop_id, GValue *valu
 
 	switch (prop_id) {
 	case PROP_CERTIFICATE:
-		g_value_set_object (value, self->pv->certificate);
+		g_value_set_object (value, self->pv->opt_cert);
 		break;
 	case PROP_LABEL:
-		g_value_take_string (value, calculate_label (self, NULL));
+		g_value_take_string (value, calculate_label (self));
 		break;
 	case PROP_ATTRIBUTES:
-		g_value_set_boxed (value, self->pv->attributes);
+		g_value_set_boxed (value, self->pv->opt_attrs);
 		break;
 	default:
-		G_OBJECT_WARN_INVALID_PROPERTY_ID (obj, prop_id, pspec);
+		gcr_certificate_mixin_get_property (obj, prop_id, value, pspec);
 		break;
 	}
 }
@@ -280,6 +516,12 @@ gcr_certificate_renderer_class_init (GcrCertificateRendererClass *klass)
 	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
 	GckAttributes *registered;
 
+	OID_SUBJECT_KEY_IDENTIFIER = g_quark_from_static_string ("2.5.29.14");
+	OID_BASIC_CONSTRAINTS = g_quark_from_static_string ("2.5.29.19");
+	OID_EXTENDED_KEY_USAGE = g_quark_from_static_string ("2.5.29.37");
+	OID_KEY_USAGE = g_quark_from_static_string ("2.5.29.15");
+	OID_SUBJECT_ALT_NAME = g_quark_from_static_string ("2.5.29.17");
+
 	gcr_certificate_renderer_parent_class = g_type_class_peek_parent (klass);
 	g_type_class_add_private (klass, sizeof (GcrCertificateRendererPrivate));
 
@@ -294,13 +536,30 @@ gcr_certificate_renderer_class_init (GcrCertificateRendererClass *klass)
 	 * The certificate to display. May be %NULL.
 	 */
 	g_object_class_install_property (gobject_class, PROP_CERTIFICATE,
-	           g_param_spec_object("certificate", "Certificate", "Certificate to display.",
-	                               GCR_TYPE_CERTIFICATE, G_PARAM_READWRITE));
+	           g_param_spec_object ("certificate", "Certificate", "Certificate to display.",
+	                                GCR_TYPE_CERTIFICATE, G_PARAM_READWRITE));
 
-	g_object_class_override_property (gobject_class, PROP_LABEL, "label");
-	g_object_class_override_property (gobject_class, PROP_ATTRIBUTES, "attributes");
+	/**
+	 * GcrCertificateRenderer:attributes:
+	 *
+	 * The certificate attributes to display. One of the attributes must be
+	 * a CKA_VALUE type attribute which contains a DER encoded certificate.
+	 */
+	g_object_class_install_property (gobject_class, PROP_ATTRIBUTES,
+	           g_param_spec_boxed ("attributes", "Attributes", "Certificate pkcs11 attributes",
+	                               GCK_TYPE_ATTRIBUTES, G_PARAM_READWRITE));
+
+	/**
+	 * GcrCertificateRenderer:label:
+	 *
+	 * The label to display.
+	 */
+	g_object_class_install_property (gobject_class, PROP_LABEL,
+	           g_param_spec_string ("label", "Label", "Certificate Label",
+	                                "", G_PARAM_READWRITE));
 
 	_gcr_icons_register ();
+	gcr_certificate_mixin_class_init (gobject_class);
 
 	/* Register this as a renderer which can be loaded */
 	registered = gck_attributes_new ();
@@ -310,7 +569,7 @@ gcr_certificate_renderer_class_init (GcrCertificateRendererClass *klass)
 }
 
 static void
-gcr_certificate_renderer_real_render (GcrRenderer *renderer, GcrViewer *viewer)
+gcr_certificate_renderer_render (GcrRenderer *renderer, GcrViewer *viewer)
 {
 	GcrCertificateRenderer *self;
 	gconstpointer data, value;
@@ -318,6 +577,7 @@ gcr_certificate_renderer_real_render (GcrRenderer *renderer, GcrViewer *viewer)
 	GcrDisplayView *view;
 	on_parsed_dn_args args;
 	const gchar *text;
+	GcrCertificate *cert;
 	gpointer raw;
 	gulong version;
 	guint bits, index;
@@ -325,6 +585,7 @@ gcr_certificate_renderer_real_render (GcrRenderer *renderer, GcrViewer *viewer)
 	GNode *asn;
 	GQuark oid;
 	GDate date;
+	GIcon *icon;
 
 	self = GCR_CERTIFICATE_RENDERER (renderer);
 
@@ -338,19 +599,20 @@ gcr_certificate_renderer_real_render (GcrRenderer *renderer, GcrViewer *viewer)
 	}
 
 	_gcr_display_view_clear (view, renderer);
+	cert = GCR_CERTIFICATE (self);
 
-	if (!self->pv->certificate)
+	data = gcr_certificate_get_der_data (cert, &n_data);
+	if (!data)
 		return;
 
-	_gcr_display_view_set_stock_image (view, GCR_RENDERER (self), GCR_ICON_CERTIFICATE);
-
-	data = gcr_certificate_get_der_data (self->pv->certificate, &n_data);
-	g_return_if_fail (data);
+	icon = gcr_certificate_get_icon (cert);
+	_gcr_display_view_set_icon (view, GCR_RENDERER (self), icon);
+	g_object_unref (icon);
 
 	asn = egg_asn1x_create_and_decode (pkix_asn1_tab, "Certificate", data, n_data);
 	g_return_if_fail (asn);
 
-	display = calculate_label (self, asn);
+	display = calculate_label (self);
 	_gcr_display_view_append_title (view, renderer, display);
 	g_free (display);
 
@@ -394,9 +656,7 @@ gcr_certificate_renderer_real_render (GcrRenderer *renderer, GcrViewer *viewer)
 
 	raw = egg_asn1x_get_integer_as_raw (egg_asn1x_node (asn, "tbsCertificate", "serialNumber", NULL), NULL, &n_raw);
 	g_return_if_fail (raw);
-	display = egg_hex_encode_full (raw, n_raw, TRUE, ' ', 1);
-	_gcr_display_view_append_value (view, renderer, _("Serial Number"), display, TRUE);
-	g_free (display);
+	_gcr_display_view_append_hex (view, renderer, _("Serial Number"), raw, n_raw);
 	g_free (raw);
 
 	display = g_malloc0 (128);
@@ -412,25 +672,26 @@ gcr_certificate_renderer_real_render (GcrRenderer *renderer, GcrViewer *viewer)
 	}
 	g_free (display);
 
+	/* Fingerprints */
+	_gcr_display_view_append_heading (view, renderer, _("Certificate Fingerprints"));
+
+	_gcr_display_view_append_fingerprint (view, renderer, data, n_data, "SHA1", G_CHECKSUM_SHA1);
+	_gcr_display_view_append_fingerprint (view, renderer, data, n_data, "MD5", G_CHECKSUM_MD5);
+
 	/* Signature */
-	_gcr_display_view_append_heading (view, renderer, ("Signature"));
+	_gcr_display_view_append_heading (view, renderer, _("Signature"));
 
 	oid = egg_asn1x_get_oid_as_quark (egg_asn1x_node (asn, "signatureAlgorithm", "algorithm", NULL));
 	text = egg_oid_get_description (oid);
 	_gcr_display_view_append_value (view, renderer, _("Signature Algorithm"), text, FALSE);
 
 	value = egg_asn1x_get_raw_element (egg_asn1x_node (asn, "signatureAlgorithm", "parameters", NULL), &n_value);
-	if (value && n_value) {
-		display = egg_hex_encode_full (value, n_value, TRUE, ' ', 1);
-		_gcr_display_view_append_value (view, renderer, _("Signature Parameters"), display, TRUE);
-		g_free (display);
-	}
+	if (value && n_value)
+		_gcr_display_view_append_hex (view, renderer, _("Signature Parameters"), value, n_value);
 
 	raw = egg_asn1x_get_bits_as_raw (egg_asn1x_node (asn, "signature", NULL), NULL, &bits);
 	g_return_if_fail (raw);
-	display = egg_hex_encode_full (raw, bits / 8, TRUE, ' ', 1);
-	_gcr_display_view_append_value (view, renderer, _("Signature"), display, TRUE);
-	g_free (display);
+	_gcr_display_view_append_hex (view, renderer, _("Signature"), raw, bits / 8);
 	g_free (raw);
 
 	/* Public Key Info */
@@ -443,33 +704,28 @@ gcr_certificate_renderer_real_render (GcrRenderer *renderer, GcrViewer *viewer)
 
 	value = egg_asn1x_get_raw_element (egg_asn1x_node (asn, "tbsCertificate", "subjectPublicKeyInfo",
 	                                                   "algorithm", "parameters", NULL), &n_value);
-	if (value && n_value) {
-		display = egg_hex_encode_full (value, n_value, TRUE, ' ', 1);
-		_gcr_display_view_append_value (view, renderer, _("Key Parameters"), display, TRUE);
-		g_free (display);
-	}
+	if (value && n_value)
+		_gcr_display_view_append_hex (view, renderer, _("Key Parameters"), value, n_value);
 
-	bits = gcr_certificate_get_key_size (self->pv->certificate);
+	bits = gcr_certificate_get_key_size (cert);
 	if (bits > 0) {
 		display = g_strdup_printf ("%u", bits);
 		_gcr_display_view_append_value (view, renderer, _("Key Size"), display, FALSE);
 		g_free (display);
 	}
 
+	value = egg_asn1x_get_raw_element (egg_asn1x_node (asn, "tbsCertificate",
+	                                                   "subjectPublicKeyInfo", NULL), &n_value);
+	raw = _gcr_fingerprint_from_subject_public_key_info (value, n_value, G_CHECKSUM_SHA1, &n_raw);
+	_gcr_display_view_append_hex (view, renderer, _("Key SHA1 Fingerprint"), raw, n_raw);
+	g_free (raw);
+
 	raw = egg_asn1x_get_bits_as_raw (egg_asn1x_node (asn, "tbsCertificate", "subjectPublicKeyInfo",
 	                                                 "subjectPublicKey", NULL), NULL, &bits);
 	g_return_if_fail (raw);
-	display = egg_hex_encode_full (raw, bits / 8, TRUE, ' ', 1);
-	_gcr_display_view_append_value (view, renderer, _("Public Key"), display, TRUE);
-	g_free (display);
+	_gcr_display_view_append_hex (view, renderer, _("Public Key"), raw, bits / 8);
 	g_free (raw);
 
-	/* Fingerprints */
-	_gcr_display_view_append_heading (view, renderer, _("Fingerprints"));
-
-	_gcr_display_view_append_fingerprint (view, renderer, data, n_data, "SHA1", G_CHECKSUM_SHA1);
-	_gcr_display_view_append_fingerprint (view, renderer, data, n_data, "MD5", G_CHECKSUM_MD5);
-
 	/* Extensions */
 	for (index = 1; TRUE; ++index) {
 		if (!append_extension (self, view, asn, data, n_data, index))
@@ -479,10 +735,55 @@ gcr_certificate_renderer_real_render (GcrRenderer *renderer, GcrViewer *viewer)
 	egg_asn1x_destroy (asn);
 }
 
+static void
+gcr_certificate_renderer_populate_popup (GcrRenderer *self, GcrViewer *viewer,
+                                         GtkMenu *menu)
+{
+	GtkWidget *item;
+
+	item = gtk_separator_menu_item_new ();
+	gtk_widget_show (item);
+	gtk_menu_shell_prepend (GTK_MENU_SHELL (menu), item);
+
+	item = gtk_menu_item_new_with_label ("Export Certificate...");
+	gtk_widget_show (item);
+	g_signal_connect_data (item, "activate", G_CALLBACK (on_certificate_export),
+	                       g_object_ref (self), (GClosureNotify)g_object_unref, 0);
+	gtk_menu_shell_prepend (GTK_MENU_SHELL (menu), item);
+}
+
 static void
 gcr_renderer_iface_init (GcrRendererIface *iface)
 {
-	iface->render = gcr_certificate_renderer_real_render;
+	iface->populate_popup = gcr_certificate_renderer_populate_popup;
+	iface->render_view = gcr_certificate_renderer_render;
+}
+
+static gconstpointer
+gcr_certificate_renderer_get_der_data (GcrCertificate *cert, gsize *n_data)
+{
+	GcrCertificateRenderer *self = GCR_CERTIFICATE_RENDERER (cert);
+	GckAttribute *attr;
+
+	g_assert (n_data);
+
+	if (self->pv->opt_cert)
+		return gcr_certificate_get_der_data (self->pv->opt_cert, n_data);
+
+	if (self->pv->opt_attrs) {
+		attr = gck_attributes_find (self->pv->opt_attrs, CKA_VALUE);
+		g_return_val_if_fail (attr, NULL);
+		*n_data = attr->length;
+		return attr->value;
+	}
+
+	return NULL;
+}
+
+static void
+gcr_renderer_certificate_iface_init (GcrCertificateIface *iface)
+{
+	iface->get_der_data = gcr_certificate_renderer_get_der_data;
 }
 
 /* -----------------------------------------------------------------------------
@@ -504,6 +805,24 @@ gcr_certificate_renderer_new (GcrCertificate *certificate)
 	return g_object_new (GCR_TYPE_CERTIFICATE_RENDERER, "certificate", certificate, NULL);
 }
 
+/**
+ * gcr_certificate_renderer_new_for_attributes:
+ * @label: The label to display
+ * @attrs: The attributes to display
+ *
+ * Create a new certificate renderer to display the label and attributes. One
+ * of the attributes should be a CKA_VALUE type attribute containing a DER
+ * encoded certificate.
+ *
+ * Returns: A newly allocated #GcrCertificateRenderer, which should be released
+ *     with g_object_unref().
+ */
+GcrCertificateRenderer*
+gcr_certificate_renderer_new_for_attributes (const gchar *label, struct _GckAttributes *attrs)
+{
+	return g_object_new (GCR_TYPE_CERTIFICATE_RENDERER, "label", label, "attributes", attrs, NULL);
+}
+
 /**
  * gcr_certificate_renderer_get_certificate:
  * @self: The renderer
@@ -518,7 +837,9 @@ GcrCertificate*
 gcr_certificate_renderer_get_certificate (GcrCertificateRenderer *self)
 {
 	g_return_val_if_fail (GCR_IS_CERTIFICATE_RENDERER (self), NULL);
-	return self->pv->certificate;
+	if (self->pv->opt_cert)
+		return self->pv->opt_cert;
+	return GCR_CERTIFICATE (self);
 }
 
 /**
@@ -533,11 +854,16 @@ gcr_certificate_renderer_set_certificate (GcrCertificateRenderer *self, GcrCerti
 {
 	g_return_if_fail (GCR_IS_CERTIFICATE_RENDERER (self));
 
-	if (self->pv->certificate)
-		g_object_unref (self->pv->certificate);
-	self->pv->certificate = certificate;
-	if (self->pv->certificate)
-		g_object_ref (self->pv->certificate);
+	if (self->pv->opt_cert)
+		g_object_unref (self->pv->opt_cert);
+	self->pv->opt_cert = certificate;
+	if (self->pv->opt_cert)
+		g_object_ref (self->pv->opt_cert);
+
+	if (self->pv->opt_attrs) {
+		gck_attributes_unref (self->pv->opt_attrs);
+		self->pv->opt_attrs = NULL;
+	}
 
 	gcr_renderer_emit_data_changed (GCR_RENDERER (self));
 	g_object_notify (G_OBJECT (self), "certificate");
@@ -555,7 +881,7 @@ GckAttributes*
 gcr_certificate_renderer_get_attributes (GcrCertificateRenderer *self)
 {
 	g_return_val_if_fail (GCR_IS_CERTIFICATE_RENDERER (self), NULL);
-	return self->pv->attributes;
+	return self->pv->opt_attrs;
 }
 
 /**
@@ -569,33 +895,21 @@ gcr_certificate_renderer_get_attributes (GcrCertificateRenderer *self)
 void
 gcr_certificate_renderer_set_attributes (GcrCertificateRenderer *self, GckAttributes *attrs)
 {
-	GcrCertificate *cert;
-	GckAttribute *attr;
-	gboolean emit = TRUE;
-
 	g_return_if_fail (GCR_IS_CERTIFICATE_RENDERER (self));
 
-	gck_attributes_unref (self->pv->attributes);
-	self->pv->attributes = attrs;\
+	gck_attributes_unref (self->pv->opt_attrs);
+	self->pv->opt_attrs = attrs;
 
-	if (self->pv->attributes) {
-		gck_attributes_ref (self->pv->attributes);
-		attr = gck_attributes_find (self->pv->attributes, CKA_VALUE);
-		if (attr) {
-			/* Create a new certificate object refferring to same memory */
-			cert = gcr_simple_certificate_new_static (attr->value, attr->length);
-			g_object_set_data_full (G_OBJECT (cert), "attributes",
-			                        gck_attributes_ref (self->pv->attributes),
-			                        (GDestroyNotify)gck_attributes_unref);
-			gcr_certificate_renderer_set_certificate (self, cert);
-			g_object_unref (cert);
-			emit = FALSE;
-		} else {
-			gcr_certificate_renderer_set_certificate (self, NULL);
-		}
+	if (self->pv->opt_attrs)
+		gck_attributes_ref (self->pv->opt_attrs);
+
+	if (self->pv->opt_cert) {
+		g_object_unref (self->pv->opt_cert);
+		g_object_notify (G_OBJECT (self), "certificate");
+		self->pv->opt_cert = NULL;
 	}
 
-	if (emit)
-		gcr_renderer_emit_data_changed (GCR_RENDERER (self));
+	gcr_renderer_emit_data_changed (GCR_RENDERER (self));
+	g_object_notify (G_OBJECT (self), "attributes");
 
 }
