@@ -27,6 +27,7 @@
 #include "egg/egg-hex.h"
 
 #include <gdk/gdk.h>
+#include <glib/gi18n-lib.h>
 
 static void _gcr_display_view_viewer_iface (GcrViewerIface *iface);
 
@@ -62,6 +63,7 @@ struct _GcrDisplayViewPrivate {
 	GtkTextTag *content_tag;
 	GtkTextTag *heading_tag;
 	GtkTextTag *monospace_tag;
+	GcrDisplayItem *current_item;
 
 	gboolean have_measurements;
 	gint minimal_width;
@@ -74,8 +76,6 @@ struct _GcrDisplayViewPrivate {
  * INTERNAL
  */
 
-#if GTK_CHECK_VERSION (3,0,0)
-
 static void
 ensure_measurements (GcrDisplayView *self)
 {
@@ -134,8 +134,6 @@ ensure_measurements (GcrDisplayView *self)
 	self->pv->have_measurements = TRUE;
 }
 
-#endif /* GTK+ 3 */
-
 static void
 recalculate_and_resize (GcrDisplayView *self)
 {
@@ -196,11 +194,7 @@ on_expander_realize (GtkWidget *widget, gpointer user_data)
 {
 	GdkCursor *cursor = gdk_cursor_new (GDK_ARROW);
 	gdk_window_set_cursor (gtk_widget_get_window (widget), cursor);
-#if GTK_CHECK_VERSION (2,91,7)
 	g_object_unref (cursor);
-#else
-	gdk_cursor_unref (cursor);
-#endif
 }
 
 static void
@@ -209,14 +203,13 @@ on_expander_expanded (GObject *object, GParamSpec *param_spec, gpointer user_dat
 	GtkExpander *expander = GTK_EXPANDER (object);
 	GcrDisplayItem *item = user_data;
 	item->expanded = gtk_expander_get_expanded (expander);
-	gcr_renderer_render (item->renderer, GCR_VIEWER (item->display_view));
+	gcr_renderer_render_view (item->renderer, GCR_VIEWER (item->display_view));
 	recalculate_and_resize (item->display_view);
 }
 
 static void
 style_display_item (GtkWidget *widget, GcrDisplayItem *item)
 {
-#if GTK_CHECK_VERSION (2,91,6)
 	GtkStyleContext *style;
 	GdkRGBA color;
 
@@ -229,10 +222,6 @@ style_display_item (GtkWidget *widget, GcrDisplayItem *item)
 	gtk_style_context_restore (style);
 
 	gtk_widget_override_background_color (item->details_widget, GTK_STATE_NORMAL, &color);
-#else
-	GtkStyle *style = gtk_widget_get_style (widget);
-	gtk_widget_modify_bg (item->details_widget, GTK_STATE_NORMAL, &style->base[GTK_STATE_NORMAL]);
-#endif
 }
 
 static GcrDisplayItem*
@@ -244,6 +233,7 @@ create_display_item (GcrDisplayView *self, GcrRenderer *renderer)
 	GtkWidget *widget;
 	GtkWidget *label;
 	GtkWidget *alignment;
+	gchar *text;
 
 	item = g_new0 (GcrDisplayItem, 1);
 	item->display_view = self;
@@ -277,10 +267,12 @@ create_display_item (GcrDisplayView *self, GcrRenderer *renderer)
 
 	widget = gtk_expander_new_with_mnemonic ("");
 	label = gtk_expander_get_label_widget (GTK_EXPANDER (widget));
-	gtk_label_set_markup_with_mnemonic (GTK_LABEL (label), "<b>_Details</b>");
+	text = g_strdup_printf ("<b>%s</b>", _("_Details"));
+	gtk_label_set_markup_with_mnemonic (GTK_LABEL (label), text);
 	g_signal_connect (widget, "notify::expanded", G_CALLBACK (on_expander_expanded), item);
 	g_signal_connect (widget, "realize", G_CALLBACK (on_expander_realize), NULL);
 	item->expanded = gtk_expander_get_expanded (GTK_EXPANDER (widget));
+	g_free (text);
 
 	alignment = gtk_alignment_new (0.5, 0.5, 0.5, 0.5);
 	gtk_alignment_set_padding (GTK_ALIGNMENT (alignment), 6, 9, 0, 0);
@@ -347,11 +339,34 @@ lookup_display_item (GcrDisplayView *self, GcrRenderer *renderer)
 	return item;
 }
 
+static GcrDisplayItem*
+find_item_at_iter (GcrDisplayView *self, GtkTextIter *iter)
+{
+	GHashTableIter hi;
+	GcrDisplayItem *item;
+	gpointer value;
+	GtkTextIter start, end;
+
+	g_hash_table_iter_init (&hi, self->pv->items);
+	while (g_hash_table_iter_next (&hi, NULL, &value)) {
+		item = value;
+
+		gtk_text_buffer_get_iter_at_mark (self->pv->buffer, &start, item->beginning);
+		gtk_text_buffer_get_iter_at_mark (self->pv->buffer, &end, item->ending);
+
+		if (gtk_text_iter_compare (iter, &start) >= 0 &&
+		    gtk_text_iter_compare (iter, &end) < 0)
+			return item;
+	}
+
+	return NULL;
+}
+
 static void
 on_renderer_data_changed (GcrRenderer *renderer, GcrViewer *self)
 {
 	/* Just ask the renderer to render itself on us */
-	gcr_renderer_render (renderer, self);
+	gcr_renderer_render_view (renderer, self);
 }
 
 static void
@@ -514,7 +529,31 @@ _gcr_display_view_realize (GtkWidget *widget)
 		style_display_item (widget, value);
 }
 
-#if GTK_CHECK_VERSION (2,91,0)
+static gboolean
+_gcr_display_view_button_press_event (GtkWidget *widget, GdkEventButton *event)
+{
+	GtkTextView *text_view = GTK_TEXT_VIEW (widget);
+	GcrDisplayView *self = GCR_DISPLAY_VIEW (widget);
+	GcrDisplayItem *item;
+	gboolean handled = FALSE;
+	GtkTextIter iter;
+	gint x, y;
+
+	if (GTK_WIDGET_CLASS (_gcr_display_view_parent_class)->button_press_event)
+		handled = GTK_WIDGET_CLASS (_gcr_display_view_parent_class)->button_press_event (
+				widget, event);
+
+	if (event->window == gtk_text_view_get_window (text_view, GTK_TEXT_WINDOW_TEXT)) {
+		gtk_text_view_window_to_buffer_coords (text_view, GTK_TEXT_WINDOW_TEXT,
+		                                       event->x, event->y, &x, &y);
+		gtk_text_view_get_iter_at_location (text_view, &iter, x, y);
+
+		item = find_item_at_iter (self, &iter);
+		self->pv->current_item = item;
+	}
+
+	return handled;
+}
 
 static gboolean
 _gcr_display_view_draw (GtkWidget *widget, cairo_t *cr)
@@ -533,33 +572,6 @@ _gcr_display_view_draw (GtkWidget *widget, cairo_t *cr)
 	return handled;
 }
 
-#else /* GTK 2.x */
-
-static gboolean
-_gcr_display_view_expose_event (GtkWidget *widget, GdkEventExpose *event)
-{
-	GcrDisplayView *self = GCR_DISPLAY_VIEW (widget);
-	gboolean handled = FALSE;
-	cairo_t *cr;
-
-	/* Have GtkTextView draw the text first. */
-	if (GTK_WIDGET_CLASS (_gcr_display_view_parent_class)->expose_event)
-		handled = GTK_WIDGET_CLASS (_gcr_display_view_parent_class)->expose_event (widget, event);
-
-	/* Render the pixbuf if it's available */
-	if (event->window == gtk_text_view_get_window (GTK_TEXT_VIEW (widget), GTK_TEXT_WINDOW_TEXT)) {
-		cr = gdk_cairo_create (event->window);
-		paint_widget_icons (GCR_DISPLAY_VIEW (self), cr);
-		cairo_destroy (cr);
-	}
-
-	return handled;
-}
-
-#endif /* GTK 2.x */
-
-#if GTK_CHECK_VERSION (3,0,0)
-
 static void
 _gcr_display_get_preferred_height (GtkWidget *widget, gint *minimal_height,
                                    gint *natural_height)
@@ -580,13 +592,26 @@ _gcr_display_get_preferred_width (GtkWidget *widget, gint *minimal_width,
 	*natural_width = self->pv->natural_width;
 }
 
-#endif /* GTK 3.x */
+static void
+_gcr_display_view_populate_popup (GtkTextView *text_view, GtkMenu *menu)
+{
+	GcrDisplayView *self = GCR_DISPLAY_VIEW (text_view);
+
+	if (GTK_TEXT_VIEW_CLASS (_gcr_display_view_parent_class)->populate_popup)
+		GTK_TEXT_VIEW_CLASS (_gcr_display_view_parent_class)->populate_popup (text_view, menu);
+
+	/* Ask the current renderer to add menu items */
+	if (self->pv->current_item)
+		gcr_renderer_popuplate_popup (self->pv->current_item->renderer,
+		                              GCR_VIEWER (self), menu);
+}
 
 static void
 _gcr_display_view_class_init (GcrDisplayViewClass *klass)
 {
 	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
 	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
+	GtkTextViewClass *text_view_class = GTK_TEXT_VIEW_CLASS (klass);
 
 	_gcr_display_view_parent_class = g_type_class_peek_parent (klass);
 	g_type_class_add_private (klass, sizeof (GcrDisplayViewPrivate));
@@ -596,17 +621,12 @@ _gcr_display_view_class_init (GcrDisplayViewClass *klass)
 	gobject_class->finalize = _gcr_display_view_finalize;
 
 	widget_class->realize = _gcr_display_view_realize;
-
-#if GTK_CHECK_VERSION (3,0,0)
+	widget_class->button_press_event = _gcr_display_view_button_press_event;
 	widget_class->get_preferred_height = _gcr_display_get_preferred_height;
 	widget_class->get_preferred_width = _gcr_display_get_preferred_width;
-#endif
-
-#if GTK_CHECK_VERSION (2,91,0)
 	widget_class->draw = _gcr_display_view_draw;
-#else
-	widget_class->expose_event = _gcr_display_view_expose_event;
-#endif
+
+	text_view_class->populate_popup = _gcr_display_view_populate_popup;
 }
 
 static void
@@ -619,7 +639,7 @@ _gcr_display_view_real_add_renderer (GcrViewer *viewer, GcrRenderer *renderer)
 	g_ptr_array_add (self->pv->renderers, g_object_ref (renderer));
 	g_hash_table_insert (self->pv->items, renderer, item);
 
-	gcr_renderer_render (renderer, viewer);
+	gcr_renderer_render_view (renderer, viewer);
 	item->data_changed_id = g_signal_connect (renderer, "data-changed",
 	                                          G_CALLBACK (on_renderer_data_changed), self);
 }
@@ -816,6 +836,17 @@ _gcr_display_view_append_value (GcrDisplayView *self, GcrRenderer *renderer, con
 	g_free (text);
 }
 
+void
+_gcr_display_view_append_hex (GcrDisplayView *self, GcrRenderer *renderer,
+                              const gchar *field, gconstpointer value, gsize n_value)
+{
+	gchar *display;
+
+	display = egg_hex_encode_full (value, n_value, TRUE, ' ', 1);
+	_gcr_display_view_append_value (self, renderer, field, display, TRUE);
+	g_free (display);
+}
+
 void
 _gcr_display_view_append_title (GcrDisplayView *self, GcrRenderer *renderer, const gchar *title)
 {
@@ -867,7 +898,6 @@ _gcr_display_view_append_fingerprint (GcrDisplayView *self, GcrRenderer *rendere
 	GChecksum *checksum;
 	guint8 *buffer;
 	gsize n_buffer;
-	gchar *display;
 
 	g_return_if_fail (GCR_IS_DISPLAY_VIEW (self));
 
@@ -882,18 +912,20 @@ _gcr_display_view_append_fingerprint (GcrDisplayView *self, GcrRenderer *rendere
 	g_checksum_get_digest (checksum, buffer, &n_buffer);
 	g_checksum_free (checksum);
 
-	display = egg_hex_encode_full (buffer, n_buffer, TRUE, ' ', 1);
-	_gcr_display_view_append_value (self, renderer, name, display, TRUE);
-	g_free (display);
+	_gcr_display_view_append_hex (self, renderer, name, buffer, n_buffer);
 
 	g_free (buffer);
 }
 
 void
-_gcr_display_view_set_stock_image (GcrDisplayView *self, GcrRenderer *renderer,
-                                   const gchar *stock_id)
+_gcr_display_view_set_icon (GcrDisplayView *self, GcrRenderer *renderer, GIcon *icon)
 {
 	GcrDisplayItem *item;
+	GdkScreen *screen;
+	GtkIconTheme *icon_theme;
+	GtkSettings *settings;
+	gint width, height;
+	GtkIconInfo *info;
 
 	g_return_if_fail (GCR_IS_DISPLAY_VIEW (self));
 	item = lookup_display_item (self, renderer);
@@ -901,12 +933,24 @@ _gcr_display_view_set_stock_image (GcrDisplayView *self, GcrRenderer *renderer,
 
 	if (item->pixbuf)
 		g_object_unref (item->pixbuf);
-	if (stock_id)
-#if GTK_CHECK_VERSION (2,91,7)
-		item->pixbuf = gtk_widget_render_icon_pixbuf (GTK_WIDGET (self), stock_id, GTK_ICON_SIZE_DIALOG);
-#else
-		item->pixbuf = gtk_widget_render_icon (GTK_WIDGET (self), stock_id, GTK_ICON_SIZE_DIALOG, NULL);
-#endif
-	else
-		item->pixbuf = NULL;
+	item->pixbuf = NULL;
+
+	if (!icon)
+		return;
+
+	screen = gtk_widget_get_screen (GTK_WIDGET (self));
+	icon_theme = gtk_icon_theme_get_for_screen (screen);
+	settings = gtk_settings_get_for_screen (screen);
+
+	if (!gtk_icon_size_lookup_for_settings (settings, GTK_ICON_SIZE_DIALOG, &width, &height))
+		g_return_if_reached ();
+
+	info = gtk_icon_theme_lookup_by_gicon (icon_theme, icon, MIN (width, height),
+	                                       GTK_ICON_LOOKUP_USE_BUILTIN);
+
+	if (info) {
+		GtkStyleContext *style = gtk_widget_get_style_context (GTK_WIDGET (self));
+		item->pixbuf = gtk_icon_info_load_symbolic_for_context (info, style, FALSE, NULL);
+		gtk_icon_info_free (info);
+	}
 }
