@@ -24,6 +24,7 @@
 
 #include <glib-object.h>
 
+#include "gcr-column.h"
 #include "gcr-types.h"
 #include "gcr-viewer.h"
 
@@ -43,9 +44,13 @@ struct _GcrRendererIface {
 	void (*data_changed) (GcrRenderer *self);
 
 	/* virtual */
-	void (*render) (GcrRenderer *self, GcrViewer *viewer);
+	void (*render_view) (GcrRenderer *self, GcrViewer *viewer);
 
+	void (*populate_popup) (GcrRenderer *self, GcrViewer *viewer, GtkMenu *menu);
+
+	/*< private >*/
 	gpointer dummy1;
+
 	gpointer dummy2;
 	gpointer dummy3;
 	gpointer dummy4;
@@ -55,17 +60,21 @@ struct _GcrRendererIface {
 
 };
 
-GType                  gcr_renderer_get_type                      (void) G_GNUC_CONST;
+GType                     gcr_renderer_get_type                   (void) G_GNUC_CONST;
 
-void                   gcr_renderer_render                        (GcrRenderer *self,
+void                      gcr_renderer_render_view                (GcrRenderer *self,
                                                                    GcrViewer *viewer);
 
-void                   gcr_renderer_emit_data_changed             (GcrRenderer *self);
+void                      gcr_renderer_popuplate_popup            (GcrRenderer *self,
+                                                                   GcrViewer *viewer,
+                                                                   GtkMenu *menu);
 
-GcrRenderer*           gcr_renderer_create                        (const gchar *label,
+void                      gcr_renderer_emit_data_changed          (GcrRenderer *self);
+
+GcrRenderer*              gcr_renderer_create                     (const gchar *label,
                                                                    GckAttributes *attrs);
 
-void                   gcr_renderer_register                      (GType renderer_type,
+void                      gcr_renderer_register                   (GType renderer_type,
                                                                    GckAttributes *attrs);
 
 G_END_DECLS
