@@ -22,7 +22,6 @@
 #include "config.h"
 
 #include "pkcs11/pkcs11.h"
-#include "pkcs11/pkcs11g.h"
 #include "pkcs11/pkcs11i.h"
 
 #include "gkm-aes-key.h"
@@ -31,6 +30,7 @@
 #include "gkm-certificate.h"
 #include "gkm-credential.h"
 #include "gkm-factory.h"
+#include "gkm-generic-key.h"
 #include "gkm-manager.h"
 #include "gkm-memory-store.h"
 #include "gkm-module.h"
@@ -161,9 +161,15 @@ static const MechanismAndInfo mechanism_list[] = {
 
 	/*
 	 * CKM_DH_PKCS_DERIVE
-	 * For DH derivation the min and max are sizes of prime in bits.
+	 * For DH derivation the min and max are sizes of output key in bytes.
 	 */
-	{ CKM_DH_PKCS_DERIVE, { 768, 8192, CKF_DERIVE } },
+	{ CKM_DH_PKCS_DERIVE, { 1, 255, CKF_DERIVE } },
+
+	/*
+	 * CKM_G_HKDF_DERIVE
+	 * For HKDF derivation the min and max are sizes of prime in bits.
+	 */
+	{ CKM_G_HKDF_SHA256_DERIVE, { 768, 8192, CKF_DERIVE } },
 
 	/*
 	 * CKM_AES_CBC_PAD
@@ -295,6 +301,20 @@ unregister_apartment (GkmModule *self, Apartment *apt)
 	g_assert (apt);
 	g_assert (GKM_IS_MODULE (self));
 
+	switch (apt->logged_in) {
+	case CKU_NONE:
+		break;
+	case CKU_USER:
+		gkm_module_logout_user (self, apt->apt_id);
+		break;
+	case CKU_SO:
+		gkm_module_logout_so (self, apt->apt_id);
+		break;
+	default:
+		g_return_if_reached ();
+		break;
+	}
+
 	if (!g_hash_table_remove (self->pv->apartments_by_id, &(apt->apt_id)))
 		g_assert_not_reached ();
 }
@@ -599,6 +619,7 @@ gkm_module_init (GkmModule *self)
 	gkm_module_register_factory (self, GKM_FACTORY_AES_KEY);
 	gkm_module_register_factory (self, GKM_FACTORY_CERTIFICATE);
 	gkm_module_register_factory (self, GKM_FACTORY_CREDENTIAL);
+	gkm_module_register_factory (self, GKM_FACTORY_GENERIC_KEY);
 	gkm_module_register_factory (self, GKM_FACTORY_NULL_KEY);
 	gkm_module_register_factory (self, GKM_FACTORY_DH_PRIVATE_KEY);
 	gkm_module_register_factory (self, GKM_FACTORY_PRIVATE_XSA_KEY);
@@ -1318,8 +1339,10 @@ gkm_module_C_Login (GkmModule *self, CK_SESSION_HANDLE handle, CK_USER_TYPE user
 	apt = lookup_apartment (self, apt_id);
 	g_return_val_if_fail (apt, CKR_GENERAL_ERROR);
 
-	if (apt->logged_in != CKU_NONE)
+	if (apt->logged_in == user_type)
 		return CKR_USER_ALREADY_LOGGED_IN;
+	if (apt->logged_in != CKU_NONE)
+		return CKR_USER_ANOTHER_ALREADY_LOGGED_IN;
 
 	if (user_type == CKU_SO) {
 
